import logging
import tokenize
import warnings

from ._internal import _Record


log = logging.getLogger(__name__)


known_dynamic_loaders = [
    '__import__',
    'import_module',
    'run_module',
    'importorskip',  # pytest helper
]


class Module(_Record):
    """Record tracking modules seen in source code"""

    def __init__(self, module_name):
        super().__init__(module_name)
        self.found_import_stmt = False
        self.found_import_fun = False


class ModuleInspection(dict):
    """Inspect usage of all top-level modules imported by a project"""

    def __missing__(self, key):
        self[key] = Module(key)
        return self[key]

    def inspect_imports(self, project_files):
        """Collect all modules found in the given files"""
        for pfile in project_files:
            try:
                self.find_imports_from_token(pfile)
            except tokenize.TokenError:
                warnings.warn(f'Found {pfile} but cannot parse it.')

    def find_imports_from_token(self, source_module):
        """Return all top level modules that are imported by `source_module`"""
        # TODO:tokens.line
        def add_package_from_statement(token):
            if token.exact_type in (tokenize.DOT, tokenize.ELLIPSIS):
                # don't record relative imports
                return
            assert token.type == tokenize.NAME, 'illegal syntax'
            self[token.string].found_import_stmt = True
            log.debug('Module %s was found imported in %s', token.string, source_module)

        def add_package_from_function(token):
            value = token.string
            while value[0].lower() in ('r', 'b', 'u', 'f'):
                if value[0].lower() == 'f':
                    return
                value = value[1:]
            if (
                len(value) >= 6
                and value[0] == value[1] == value[2] == value[-1] == value[-2] == value[-3]
            ):
                value = value[3:-3]
            else:
                value = value[1:-1]
            if value.startswith('.'):
                # don't record relative imports
                return
            self[value].found_import_fun = True
            log.debug('Module %s was found dynamically imported in %s', token.string, source_module)

        with tokenize.open(source_module) as stream:
            try:
                tokens = tokenize.generate_tokens(stream.readline)
                for token in tokens:
                    if token.type == tokenize.NAME and token.string in (
                        'raise',
                        'yield',
                    ):
                        # consume any use of yield keyword not part of an import statement
                        while not (
                            token.type == tokenize.NEWLINE or token.exact_type == tokenize.SEMI
                        ):
                            token = next(tokens)
                    if token.type == tokenize.NAME and token.string == 'import':
                        token = next(tokens)
                        add_package_from_statement(token)
                        while not (
                            token.type == tokenize.NEWLINE or token.exact_type == tokenize.SEMI
                        ):
                            token = next(tokens)
                            if token.exact_type == tokenize.COMMA:
                                token = next(tokens)
                                add_package_from_statement(token)
                    if token.type == tokenize.NAME and token.string == 'from':
                        token = next(tokens)
                        if token.exact_type == tokenize.LPAR:
                            token = next(tokens)
                        add_package_from_statement(token)
                        while not (
                            token.type == tokenize.NEWLINE or token.exact_type == tokenize.SEMI
                        ):
                            # multiple top level packages cannot be imported under a single 'from'
                            # but there may be an 'import' keyword that shouldn't go to the next if
                            token = next(tokens)
                    if token.type == tokenize.NAME and token.string in (known_dynamic_loaders):
                        try:
                            token = next(tokens)
                            if token.exact_type == tokenize.LPAR:
                                token = next(tokens)
                                while not (
                                    token.type == tokenize.NEWLINE
                                    or token.exact_type == tokenize.SEMI
                                ):
                                    if token.type == tokenize.STRING:
                                        followon_token = next(tokens)
                                        if not (
                                            followon_token.type == tokenize.OP
                                            and followon_token.exact_type
                                            not in (tokenize.COMMA, tokenize.RPAR)
                                        ):
                                            add_package_from_function(token)
                                            break
                                        # else not a constant string literal
                                    token = next(tokens)
                        except StopIteration:
                            # this is not necessarily a SyntaxError, as these are not keywords
                            pass
            except (AssertionError, StopIteration) as err:
                # If StopIteration is raised, this file contains illegal syntax
                # This would cause a SyntaxError if run, but caller is already
                # expecting TokenError
                raise tokenize.TokenError from err



import os
from argparse import HelpFormatter

from bonded.settings import CLISettings


def update_readme():
    class DocumentationFormatter(HelpFormatter):
        def __init__(self, prog, indent_increment=2, max_help_position=24, width=None):
            super().__init__(
                'bonded',
                indent_increment=indent_increment,
                max_help_position=max_help_position,
                width=80,
            )

    CLISettings.formatter_class = DocumentationFormatter
    help_txt = CLISettings.format_help()
    with open('README.md', 'r') as readme_read, open('README.new', 'w') as readme_write:
        for line in readme_read:
            readme_write.write(line)
            if line.strip() == '<!-- replace start -->':
                readme_write.write('```\n')
                readme_write.write(help_txt)
                readme_write.write('```\n')
                while line.strip() != '<!-- replace end -->':
                    line = next(readme_read)
                readme_write.write(line)
    os.rename('README.new', 'README.md')


if __name__ == '__main__':
    update_readme()



import os

import bonded.settings
from bonded.settings import CLISettings, Settings


def test_no_preferences():
    """all settings have a default"""
    Settings()


def test_excludes():
    """excludes are rewritten to be unanchored, when appropriate"""
    assert Settings().exclude == set()
    assert Settings(exclude={'/dev'}).exclude == {'/dev'}
    assert Settings(exclude={'/dev/'}).exclude == {'/dev/**'}
    assert Settings(exclude={'dev'}).exclude == {'**/dev'}
    assert Settings(exclude={'dev/'}).exclude == {'**/dev/**'}


def test_project_modules(monkeypatch):
    with monkeypatch.context() as m:
        # disable any post_init locating of modules
        m.setattr(os.path, 'isfile', lambda _a: True)
        m.setattr(os, 'listdir', lambda _a: [])
        m.setattr(os, 'walk', lambda _a: [])
        assert Settings().project_modules == set()
        assert Settings(project_modules={'foo', 'bar'}).project_modules == {'foo', 'bar'}

    with monkeypatch.context() as m:
        # single file is being checked
        m.setattr(os.path, 'isfile', lambda _a: True)
        assert Settings(search_path='test.py').project_modules == {'test'}
        assert Settings(search_path='test.md').project_modules == set()
        assert Settings(project_modules={'foo', 'bar'}, search_path='test.py').project_modules == {
            'foo',
            'bar',
            'test',
        }

    with monkeypatch.context() as m:
        # single module is being checked
        m.setattr(os.path, 'isfile', lambda _a: False)
        m.setattr(os, 'listdir', lambda _a: ['__init__.py'])
        assert Settings(search_path='test').project_modules == {'test'}
        assert Settings(project_modules={'foo', 'bar'}, search_path='test').project_modules == {
            'foo',
            'bar',
            'test',
        }

    with monkeypatch.context() as m:
        # possibly multiple packages or modules
        m.setattr(os.path, 'isfile', lambda _a: False)
        m.setattr(os, 'listdir', lambda _a: [])
        m.setattr(os, 'walk', lambda _a: [])
        assert Settings(search_path='test').project_modules == set()

        def test_walk(_):
            yield ('test', ['spam'], ['README', 'build.py'])
            yield (
                'test/spam',
                [
                    'docs',
                ],
                [
                    '__init__.py',
                ],
            )
            yield ('test/spam/docs', [], [])

        m.setattr(os, 'walk', test_walk)
        assert Settings(search_path='test').project_modules == {'spam', 'build'}

    with monkeypatch.context() as m:
        # provided pyproject.toml specifies the package name
        m.setattr(os.path, 'isfile', lambda _a: True)
        m.setattr(os, 'listdir', lambda _a: [])
        m.setattr(os, 'walk', lambda _a: [])
        m.setattr(bonded.settings, 'gather_config', lambda _a: {'project': {'name': 'test'}})
        m.setattr(bonded.settings, 'dist2pkg', {'test': ['test-py']})
        assert Settings(pyproject='pyproject.toml').project_modules == {'test-py'}
        assert Settings(
            search_path='main.py', project_modules={'lib-py'}, pyproject='pyproject.toml'
        ).project_modules == {'main', 'lib-py', 'test-py'}


def test_no_args():
    assert vars(CLISettings.parse_args([])) == {}


def test_multiple_and_many_args():
    """arguments that can have multiple values can specify them multiple
    at a time and many different times
    """
    assert vars(
        CLISettings.parse_args(
            ['--packages', 'foo', 'bar', 'baz', '--packages', 'spam', 'ham', '--packages', 'eggs']
        )
    ) == {'packages': ['foo', 'bar', 'baz', 'spam', 'ham', 'eggs']}
    assert vars(
        CLISettings.parse_args(
            [
                '--ignore-modules',
                'foo',
                'bar',
                'baz',
                '--ignore-modules',
                'spam',
                'ham',
                '--ignore-modules',
                'eggs',
            ]
        )
    ) == {'ignore_modules': ['foo', 'bar', 'baz', 'spam', 'ham', 'eggs']}
    assert vars(
        CLISettings.parse_args(
            [
                '--ignore-packages',
                'foo',
                'bar',
                'baz',
                '--ignore-packages',
                'spam',
                'ham',
                '--ignore-packages',
                'eggs',
            ]
        )
    ) == {'ignore_packages': ['foo', 'bar', 'baz', 'spam', 'ham', 'eggs']}


def test_verbosity():
    assert vars(CLISettings.parse_args(['-v'])) == {'verbose': 1}
    assert vars(CLISettings.parse_args(['-vv'])) == {'verbose': 2}
    assert vars(CLISettings.parse_args(['-vvv'])) == {'verbose': 3}
    assert vars(CLISettings.parse_args(['-vvvv'])) == {'verbose': 4}
    assert vars(CLISettings.parse_args(['-vvvvv'])) == {'verbose': 5}
    assert vars(CLISettings.parse_args(['-vvvvvvv'])) == {'verbose': 7}
    assert vars(CLISettings.parse_args(['-v', '-v', '-v'])) == {'verbose': 3}
    assert vars(CLISettings.parse_args(['-vvv', '-q'])) == {'verbose': 3, 'quiet': True}


def test_search_path():
    assert vars(CLISettings.parse_args(['path'])) == {'search_path': 'path'}
    assert vars(CLISettings.parse_args(['--', 'path'])) == {'search_path': 'path'}



from argparse import HelpFormatter

from bonded.settings import CLISettings


def test_up_to_date_options():
    class DocumentationFormatter(HelpFormatter):
        def __init__(self, prog, indent_increment=2, max_help_position=24, width=None):
            super().__init__(
                'bonded',
                indent_increment=indent_increment,
                max_help_position=max_help_position,
                width=80,
            )

    CLISettings.formatter_class = DocumentationFormatter
    help_txt = CLISettings.format_help()
    with open('README.md', 'r') as readme:
        assert (
            help_txt.strip() in readme.read()
        ), 'to update options run "python scripts/generate_documentation.py"'



_record_cache = {}


class _Record:
    """Base class to represent a project resource being tracked

    records are unique per name but may contain mutable search information
    """

    @staticmethod
    def _normalize_name(name):
        return name

    def __new__(cls, *args, **kwargs):
        name = cls._normalize_name(kwargs.get('name', args[0]))
        if (cls, name) in _record_cache:
            record = _record_cache[(cls, name)]
        else:
            record = super().__new__(cls)
            _record_cache[(cls, name)] = record
        return record

    def __init__(self, name):
        self._normalized_name = self._normalize_name(name)

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self):
        return self._normalized_name



from enum import IntEnum
from functools import cache

import tomli

from ._sys import stdlib_module_names


class Confidence(IntEnum):
    NONE = 0
    VERY_LOW = 5
    LOW = 10
    MEDIUM = 15
    HIGH = 20
    VERY_HIGH = 25
    SKIPPED = 30


class Evaluation:
    def __init__(
        self,
        packages,
        modules,
        executables,
        settings,
        stdlib_modules=None,
    ):
        self.packages = packages
        self.modules = modules
        self.executables = executables
        self.settings = settings
        self.stdlib_modules = stdlib_modules or stdlib_module_names

    def _package_platform_ignored(self, package):
        return package.markers and not any(mark.evaluate() for mark in package.markers)

    def _package_module_used(self, package):
        if not package.modules:
            return Confidence.NONE
        return max(self.evaluate_module(pkg_mod) for pkg_mod in package.modules)

    def _package_extention_used(self, package):
        extension_used = Confidence.NONE
        for extension in package.extends:
            if extension == package.name:
                continue
            extension_used = max(
                extension_used, self.evaluate_package(extension), self.evaluate_module(extension)
            )
        return extension_used

    def _package_executable_used(self, package):
        executable_used = Confidence.NONE
        for executable in package.executables:
            if executable in self.executables and self.executables[executable].found_executions:
                # TODO: check file type
                return Confidence.MEDIUM
        return executable_used

    def _package_used_anyway(self, package):
        """A series of workarounds for odd corners of the packaging world

        ANY of these that can be removed from here and worked into more general logic should be
        """
        # wheel only extends distutils, but setuptools is more frequently used
        if package.name == 'wheel' and self.evaluate_package('setuptools'):
            return Confidence.HIGH
        if 'pytest11' in package.extends and self.evaluate_package('pytest'):
            return Confidence.HIGH
        return Confidence.NONE

    @cache
    def evaluate_package(self, package):
        if package not in self.packages:
            return Confidence.NONE
        pkg = self.packages[package]
        if self._package_platform_ignored(pkg) or pkg.name in self.settings.ignore_packages:
            return Confidence.SKIPPED
        if not pkg.installed:
            return Confidence.NONE
        return max(
            self._package_module_used(pkg),
            self._package_extention_used(pkg),
            self._package_executable_used(pkg),
            self._package_used_anyway(pkg),
        )

    def _module_belongs_to_package(self, module):
        for pkg in self.packages.values():
            if module.name in pkg.modules:
                return True
            if module.name == pkg.name:
                return True
        return False

    def _module_imported(self, module):
        if module.found_import_stmt:
            return Confidence.VERY_HIGH
        if module.found_import_fun:
            return Confidence.HIGH
        return Confidence.NONE

    def _module_used_for_build(self, module):
        if not self.settings.pyproject:
            return Confidence.NONE
        with open(self.settings.pyproject, 'rb') as pyproject_file:
            pyproject = tomli.load(pyproject_file)
        build_backend = pyproject.get('build-system', {}).get('build-backend')
        backend = build_backend.split(':')[0].split('.')[0]
        if backend == module:
            return Confidence.HIGH
        return Confidence.NONE

    @cache
    def evaluate_module(self, module):
        if module not in self.modules:
            return self._module_used_for_build(module)
        mod = self.modules[module]
        if (
            mod.name in self.stdlib_modules
            or mod.name in self.settings.project_modules
            or mod.name in self.settings.ignore_modules
        ):
            return Confidence.SKIPPED
        if not self._module_belongs_to_package(mod):
            return Confidence.NONE
        return max(
            self._module_imported(mod),
            self._module_used_for_build(module),
        )

    def package_report(self):
        return {
            package for name, package in self.packages.items() if not self.evaluate_package(name)
        }

    def module_report(self):
        return {module for name, module in self.modules.items() if not self.evaluate_module(name)}

    def passes(self):
        return not bool(self.package_report() or self.module_report())


def evaluate_bonds(settings, modules, packages, executables):
    return Evaluation(
        packages,
        modules,
        executables,
        settings,
    )



from bonded.package_inspection import Package


def test_uninstalled_package():
    # illegal package name, so will never be installed
    uninstalled = Package('    ')

    assert uninstalled.package_name == '    '
    assert uninstalled.name == '    '
    assert uninstalled.modules == []
    assert uninstalled.extends == set()
    assert uninstalled.executables == set()


def test_equivilant_packages():
    bonded = Package('bonded')
    Bonded = Package('Bonded')

    assert bonded is Bonded


def test_loading_self():
    bonded = Package('bonded')

    assert bonded.package_name == 'bonded'
    assert bonded.name == 'bonded'
    assert bonded.modules == ['bonded']
    assert bonded.extends == set()
    assert bonded.executables == {'bonded'}



import logging
import os
import stat
import sysconfig
from configparser import ConfigParser
from pathlib import Path

import tomli

from packaging import requirements as pkgreq, utils as pkgutil

from ._importlib import dist2pkg, metadata as pkg_metadata
from ._internal import _Record


log = logging.getLogger(__name__)


class Package(_Record):
    """Record tracking usage of a package"""

    @staticmethod
    def _normalize_name(name):
        return pkgutil.canonicalize_name(name)

    def _executable_files(self):
        """Return executable files provided by this package"""
        for fname in pkg_metadata.files(self.package_name):
            # this is the only path that is sure to be on PATH
            # it is spelled with '/' on all platforms
            fname = str(fname)
            if fname.startswith('../../../bin/'):
                exe_name = fname[13:]
                fs_path = os.path.join(sysconfig.get_path('purelib'), fname)
                if (
                    exe_name
                    and '/' not in exe_name
                    and os.path.isfile(fs_path)
                    and (os.stat(fs_path).st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH))
                ):
                    yield exe_name

    def __init__(self, package_name):
        super().__init__(package_name)
        self.package_name = package_name
        if self.name in dist2pkg:
            self.installed = True
            self.modules = dist2pkg[self.name]
            metadata = pkg_metadata.distribution(self.name)
            self.extends = {
                ep.group.split(':')[0].split('.')[0]
                for ep in metadata.entry_points
                if ep.group != 'console_scripts'
            }
            self.executables = {
                ep.name for ep in metadata.entry_points if ep.group == 'console_scripts'
            }
            self.executables.update(exe for exe in self._executable_files())
            log.debug('Package %s was associated with modules %s', self.package_name, self.modules)
            log.debug(
                'Package %s was associated with extensions %s', self.package_name, self.extends
            )
            log.debug(
                'Package %s was associated with executables %s', self.package_name, self.executables
            )
        else:
            self.installed = False
            self.modules = []
            self.extends = set()
            self.executables = set()
            log.debug('Package %s is not installed', self.package_name)
        self.markers = []


class PackageInspection(dict):
    """Inspect usage of all package requirements by a project"""

    def __init__(self, requirements):
        for req in requirements:
            self._add_from_requirement(req)

    def __missing__(self, key):
        ckey = pkgutil.canonicalize_name(key)
        if ckey not in self:
            self[ckey] = Package(key)
        return self[ckey]

    def _add_from_requirement(self, requirement):
        parsed = pkgreq.Requirement(requirement)
        self[parsed.name]
        if parsed.marker:
            self[parsed.name].markers.append(parsed.marker)

    def update_from_pyproject(self, pyproject_toml):
        """Add all packages found as requirements in the given pyproject.toml"""
        with open(pyproject_toml, 'rb') as pyproject_file:
            pyproject = tomli.load(pyproject_file)
            project = pyproject.get('project', {})

            for dependency in project.get('dependencies', []):
                log.info(
                    'Found dependency %s in %s project.dependencies', dependency, pyproject_toml
                )
                self._add_from_requirement(dependency)

            for opt_name, optionals in project.get('optional-dependencies', {}).items():
                for optional in optionals:
                    log.info(
                        'Found dependency %s in %s project.optional-dependencies.%s',
                        optional,
                        pyproject_toml,
                        opt_name,
                    )
                    self._add_from_requirement(optional)
            for dependency in pyproject.get('build-system', {}).get('requires', []):
                log.info(
                    'Found dependency %s in %s build-system.requires', dependency, pyproject_toml
                )
                self._add_from_requirement(dependency)

    def update_from_pip_requirements(self, requirements_file):
        """Add all packages found in the given requirements file"""
        requirements_file = Path(requirements_file)
        with requirements_file.open('r') as requirements:
            for requirement in requirements:
                requirement = requirement.strip()
                if requirement.startswith('-r') or requirement.startswith('--requirements'):
                    sub_requirement = requirement.split()[1].strip()
                    self.update_from_pip_requirements(requirements_file.parent / sub_requirement)
                    continue
                if not requirement or requirement.startswith('#') or requirement.startswith('-'):
                    continue
                log.info('Found requirement %s in %s', requirement, requirements_file.name)
                self._add_from_requirement(requirement)

    def update_from_setup(self, setup_cfg):
        """Add all packages found in the given setup.cfg file"""
        setup = ConfigParser()
        setup.read(setup_cfg)
        for requirement in setup.get('options', 'install_requires', fallback='').splitlines():
            if requirement:
                log.info('Found requirement %s in %s [options]', requirement, setup_cfg)
                self._add_from_requirement(requirement)
        if 'options.extras_require' in setup:
            for requirements in setup['options.extras_require'].values():
                for requirement in requirements.splitlines():
                    if requirement:
                        log.info(
                            'Found requirement %s in %s [options.extras_require]',
                            requirement,
                            setup_cfg,
                        )
                        self._add_from_requirement(requirement)



import logging
import re
from collections import namedtuple

from ._internal import _Record


log = logging.getLogger(__name__)


def detect_file_type(file_path, first_line):
    executor = ''
    if first_line:
        if first_line.startswith(b'#!'):
            first_line = first_line[2:].strip()
            shebang = first_line.split()
            if len(shebang) > 1 and (shebang[0] == b'env' or shebang[0].endswith(b'/env')):
                shebang.pop(0)
            executor = shebang[0].rsplit(b'/')[-1]

    if file_path.suffix in [
        'py',
    ] or executor in ['python', 'python2', 'python3']:
        return 'python'
    if file_path.suffix in [
        'sh',
        'bash',
        'zsh',
        'fish',
        'xosh',
    ] or executor in ['sh', 'bash', 'zsh', 'fish', 'xosh']:
        return 'shell'
    if file_path.suffix in ['ini', 'cfg']:
        return 'ini'
    if file_path.suffix in ['yaml', 'yml']:
        return 'yaml'


_CallingFile = namedtuple('_CallingFile', ['file_name', 'file_type', 'line_number'])


class Executable(_Record):
    """Record tracking usage of an executable"""

    def __init__(self, executable_name):
        super().__init__(executable_name)
        self.found_executions = set()


class ExecutableInspection(dict):
    """Inspect usage of executables"""

    def __init__(self, keys):
        instanciated_args = ((a, Executable(a)) for a in keys)
        super().__init__(instanciated_args)

    def inspect_executables(self, project_files):
        # TODO: python -m but only after finding __main__.py
        # re.compile(fr"\bpython[\d.]*\s+-m\s+{exe}\b")
        exe_searches = {
            exe.name: re.compile(rb'\b%b\b' % exe.name.encode('utf-8')) for exe in self.values()
        }
        for project_file in project_files:
            if not project_file.is_file():
                continue
            pfile = project_file.read_bytes().splitlines()
            if not pfile:
                continue

            file_type = detect_file_type(project_file, pfile[0])

            for lineno, line in enumerate(pfile, start=1):
                for exe, search in exe_searches.items():
                    if search.search(line):
                        self[exe].found_executions.add(
                            _CallingFile(project_file.name, file_type, lineno)
                        )
                        log.debug('Found executable %s in %s:%s', exe, project_file, lineno)



import argparse
import dataclasses
import fnmatch
import os
import sys
from pathlib import Path
from typing import Optional, Set

import tomli

from ._importlib import dist2pkg, machinery


_CWD = os.getcwd()


@dataclasses.dataclass
class Settings:
    search_path: str = _CWD
    exclude: Set[str] = dataclasses.field(default_factory=set)
    packages: Set[str] = dataclasses.field(default_factory=set)
    requirements: Set[str] = dataclasses.field(default_factory=set)
    ignore_modules: Set[str] = dataclasses.field(default_factory=set)
    ignore_packages: Set[str] = dataclasses.field(default_factory=set)
    project_modules: Set[str] = dataclasses.field(default_factory=set)
    report: str = 'table'
    pyproject: Optional[str] = None
    setup: Optional[str] = None
    verbose: int = 0
    quiet: bool = False

    def __post_init__(self):
        self._unanchor_exclude()
        self._locate_project_modules()

    def _unanchor_exclude(self):
        unanchor_exclude = set()
        for i, exclude in enumerate(self.exclude):
            if not exclude.startswith(os.path.sep) and not exclude.startswith('**/'):
                exclude = f'**/{exclude}'
            if exclude.endswith('/'):
                exclude = f'{exclude}**'
            unanchor_exclude.add(exclude)
        self.exclude = unanchor_exclude

    def _locate_project_modules(self):
        if self.pyproject:
            project_name = gather_config(self.pyproject).get('project', {}).get('name', '')
            if project_name and project_name in dist2pkg:
                self.project_modules.update(dist2pkg[project_name])

        if os.path.isfile(self.search_path):
            stem, ext = os.path.splitext(os.path.basename(self.search_path))
            if ext in machinery.SOURCE_SUFFIXES:
                self.project_modules.add(stem)
        elif '__init__.py' in os.listdir(self.search_path):
            self.project_modules.add(os.path.basename(self.search_path))
        else:
            for root, dirs, files in os.walk(self.search_path):
                if any(fnmatch.fnmatch(root, exclude) for exclude in self.exclude):
                    del dirs[:]
                    continue
                if any(f == '__init__.py' for f in files):
                    self.project_modules.add(os.path.basename(root))
                    del dirs[:]
                    continue
                for f in files:
                    full_file = os.path.join(root, f)
                    if not self.exclude or not any(
                        fnmatch.fnmatch(full_file, exclude) for exclude in self.exclude
                    ):
                        stem, ext = os.path.splitext(f)
                        if ext in machinery.SOURCE_SUFFIXES:
                            self.project_modules.add(stem)

    @classmethod
    def from_interactive(cls):
        arguments = CLISettings.parse_args(sys.argv[1:])
        if not hasattr(arguments, 'pyproject'):
            pyproject = Path(getattr(arguments, 'search_path', _CWD)).resolve() / 'pyproject.toml'
            while not pyproject.is_file():
                if pyproject.parent == pyproject.parent.parent:
                    break
                pyproject = pyproject.parent.parent / 'pyproject.toml'
            else:
                arguments.pyproject = pyproject
        elif not arguments.pyproject:
            arguments.pyproject = None
        elif not os.path.isfile(arguments.pyproject):
            raise RuntimeWarning(f'Supplied --pyproject cannot be found: {arguments.pyproject}')

        settings_kwargs = {
            'search_path': _CWD,
            'exclude': set(),
            'packages': set(),
            'requirements': set(),
            'ignore_modules': set(),
            'ignore_packages': set(),
            'project_modules': set(),
            'report': 'table',
            'pyproject': None,
            'setup': None,
            'verbose': 0,
            'quiet': False,
        }
        if pyproject := getattr(arguments, 'pyproject', ''):
            pyproject_kwargs = gather_config(pyproject)
            for kw, setting in pyproject_kwargs.items():
                if isinstance(setting, list):
                    pyproject_kwargs[kw] = set(setting)
            settings_kwargs.update(pyproject_kwargs)
        arg_kwargs = vars(arguments)
        for kw, setting in arg_kwargs.items():
            if isinstance(setting, list):
                arg_kwargs[kw] = set(setting)
        settings_kwargs.update(arg_kwargs)
        return cls(**settings_kwargs)


CLISettings = argparse.ArgumentParser(argument_default=argparse.SUPPRESS)
CLISettings.add_argument(
    '--pyproject',
    help='Path to a pyproject.toml which will be searched for requirements and bonded settings',
)
CLISettings.add_argument(
    '--setup',
    help='Path to a setup.cfg which will be searched for requirements',
)
CLISettings.add_argument(
    '--packages',
    action='extend',
    nargs='+',
    help='Add a package to be checked for',
)
CLISettings.add_argument(
    '-r',
    '--requirements',
    action='append',
    help='Pip-requirements file used to specify further requirements.'
    ' Can be specified multiple times',
)
CLISettings.add_argument(
    '--ignore-modules',
    action='extend',
    nargs='+',
    help='These module will not be reported as missing a package',
)
CLISettings.add_argument(
    '--ignore-packages',
    action='extend',
    nargs='+',
    help='These packages will not be reported as unused',
)
CLISettings.add_argument(
    '--exclude',
    action='append',
    help='A glob that will exclude paths otherwise matched',
)
CLISettings.add_argument('--report', choices=['table', 'extended-table', 'line', 'none'])
CLISettings.add_argument('--verbose', '-v', action='count')
CLISettings.add_argument('--quiet', '-q', action='store_true')
CLISettings.add_argument('search_path', nargs='?')


def gather_config(pyproject):
    with open(pyproject, 'rb') as pypj:
        return tomli.load(pypj).get('tool', {}).get('bonded', {})



import fnmatch
import logging
import os
import sys
from pathlib import Path

import rich.logging

from .display import display_closing, display_report
from .evaluation import evaluate_bonds

from .executable_inspection import ExecutableInspection
from .module_inspection import ModuleInspection
from .package_inspection import PackageInspection
from .settings import Settings


log = logging.getLogger('bonded')


def setup_logging(level):
    if level < 1:
        return
    elif level > 5:
        level = 5

    lvl = [logging.CRITICAL, logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG][level - 1]
    log.addHandler(rich.logging.RichHandler(level=lvl))
    log.setLevel(lvl)


def iter_source_files(starting_dir, excludes, file_pattern):
    for root, dirs, files in os.walk(starting_dir):
        for f in fnmatch.filter(files, file_pattern):
            full = os.path.join(root, f)
            if not any(fnmatch.fnmatch(full, exclude) for exclude in excludes):
                yield Path(full)
        end = len(dirs)
        for i, d in enumerate(reversed(dirs), 1):
            full = os.path.join(root, d)
            if any(
                (fnmatch.fnmatch(full, exclude) or fnmatch.fnmatch(full + os.path.sep, exclude))
                for exclude in excludes
            ):
                del dirs[end - i]


def main():
    settings = Settings.from_interactive()

    setup_logging(settings.verbose)
    log.info('Using settings %s', settings)

    all_files = iter_source_files(settings.search_path, settings.exclude, '*')
    python_files = iter_source_files(settings.search_path, settings.exclude, '*.py')

    packages = PackageInspection(settings.packages)
    if settings.pyproject:
        packages.update_from_pyproject(settings.pyproject)
    if settings.setup:
        packages.update_from_setup(settings.setup)
    for pip_requirements in settings.requirements:
        packages.update_from_pip_requirements(pip_requirements)

    modules = ModuleInspection()
    modules.inspect_imports(python_files)

    executables = ExecutableInspection((e for p in packages.values() for e in p.executables))
    executables.inspect_executables(all_files)

    report = evaluate_bonds(settings, modules, packages, executables)

    display_report(settings, report)
    display_closing(settings, report)
    return 0 if report.passes() else 1


if __name__ == '__main__':
    sys.exit(main())



from collections import defaultdict
from importlib import machinery  # noqa: F401

import importlib_metadata as metadata
from packaging import utils as pkgutil

pkg2dist = metadata.packages_distributions()
dist2pkg = defaultdict(list)
for pkg, dists in pkg2dist.items():
    #if pkg == '..' or pkg.endswith('.dist-info'):
        #continue
    for dist in dists:
        dist = pkgutil.canonicalize_name(dist)
        dist2pkg[dist].append(pkg)
dist2pkg = dict(dist2pkg)



import pytest

from bonded.module_inspection import ModuleInspection


@pytest.fixture(autouse=True)
def module_inspection():
    return ModuleInspection()


@pytest.fixture()
def python_file(request, tmp_path, code):
    tgt = (tmp_path / request.function.__name__).with_suffix('.py')
    tgt.write_text(code)
    return tgt


@pytest.mark.parametrize(
    'code',
    [
        'import foo',
        'import foo.bar',
        'import foo as qux',
        'import foo.bar as qux',
        'from foo import bar',
        'from foo.bar import baz',
        'from foo.bar import (baz.qux)',
        'from foo import (bar as baz)',
        'from foo import bar, baz',
        'from foo import (bar, \nbaz)',
        'from foo import *',
        'yield from inner; import foo',
        'yield from inner; from foo import bar',
        'raise RuntimeError from SyntaxError; import foo',
    ],
)
def test_import_statement(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'foo' in module_inspection
    assert module_inspection['foo'].found_import_stmt


@pytest.mark.parametrize(
    'code',
    [
        'import foo, bar as baz, qux.xuq',
        'import foo; import bar as baz; import qux.xuq',
        'from foo import baz as foo; from bar import *; from qux import xuq',
    ],
)
def test_multi_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'foo' in module_inspection
    assert module_inspection['foo'].found_import_stmt
    assert 'bar' in module_inspection
    assert module_inspection['bar'].found_import_stmt
    assert 'qux' in module_inspection
    assert module_inspection['qux'].found_import_stmt


@pytest.mark.parametrize(
    'code',
    [
        'from . import foo',
        'from .. import foo',
        'from .foo import bar',
        'from .foo.bar import baz',
        'from ...foo.bar import baz',
        'yield from inner',
        'raise RuntimeError from SyntaxError',
    ],
)
def test_ignored_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert not module_inspection


@pytest.mark.parametrize(
    'code',
    [
        "__import__('foo')",
        "import_module('foo')",
        "run_module('foo')",
    ],
)
def test_dynamic_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'foo' in module_inspection
    assert module_inspection['foo'].found_import_fun


@pytest.mark.parametrize(
    'code',
    [
        '__import__(42)',
        "import_module = 'foo'",
        'run_module.__name__',
    ],
)
def test_lookalike_dynamic_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert not module_inspection


@pytest.mark.parametrize(
    'code',
    [
        "import foo; __import__('foo')",
        "from foo import bar; import_module('foo')",
        "import foo as bar; run_module('foo')",
    ],
)
def test_statement_and_dynamic_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'foo' in module_inspection
    assert module_inspection['foo'].found_import_stmt
    assert module_inspection['foo'].found_import_fun


@pytest.mark.parametrize(
    'code',
    [
        'import \\\nfoo, \\\n baz',
        'from foo import (\nbar,\n); import baz',
        'from \\\nfoo import (\\nbar,\\n); import baz',
        "__import__(\\\n'foo'\\\n) and \\n import_module('baz')",
        "__import__(\n'foo'\n) and (\n import_module('baz')\n)",
    ],
)
def test_multiline_import(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'foo' in module_inspection
    assert module_inspection['foo'].found_import_stmt or module_inspection['foo'].found_import_fun
    assert 'baz' in module_inspection
    assert module_inspection['baz'].found_import_stmt or module_inspection['baz'].found_import_fun


@pytest.mark.parametrize(
    'code',
    [
        'import sys, os',
        'from sys import path; from os import environ',
        'import sys.path, os.environ',
        'import sys as requests, os as six',
    ],
)
def test_stdlib_imports(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert 'sys' in module_inspection
    assert module_inspection['sys'].found_import_stmt
    assert 'os' in module_inspection
    assert module_inspection['os'].found_import_stmt


@pytest.mark.parametrize(
    'code',
    [
        'from .foo',
        'from ..foo import bar',
        'from ...foo import bar',
    ],
)
def test_relative_modules(python_file, module_inspection):
    module_inspection.find_imports_from_token(python_file)
    assert not module_inspection


@pytest.mark.parametrize(
    'code',
    [
        '[1, 2, 3',
        'varible = """finish this later',
        'import ',
        'from ',
        'import 42',
        'from "this" import zen',
    ],
)
def test_unparsable_python(python_file, module_inspection):
    with pytest.warns(Warning):
        module_inspection.inspect_imports([python_file])
    assert not module_inspection



import sys

from rich import print

from rich.columns import Columns
from rich.table import Table

from ._sys import stdlib_module_names
from .evaluation import Confidence


def display_closing(settings, evaluation):
    if not settings.quiet and evaluation.passes():
        print('All Good!', file=sys.stderr)


def display_report(settings, evaluation):
    format_lookup = {
        'none': (lambda _, __: None),
        'line': format_line_output,
        'table': format_table_output,
        'extended-table': format_extended_table_output,
    }
    print(format_lookup[settings.report](settings, evaluation))


def format_line_output(settings, evaluation):
    report = ''
    excess_packages = evaluation.package_report()
    if excess_packages:
        report += f"Packages: {', '.join(excess_packages)}\n"
    excess_modules = evaluation.module_report()
    if excess_modules:
        report += f"Modules: {', '.join(excess_modules)}"
    return report


def format_table_output(settings, evaluation):
    report = Columns()

    excess_packages = evaluation.package_report()
    if excess_packages:
        excess_packages_report = Table()
        excess_packages_report.add_column('Unused Package')
        excess_packages_report.add_column('Imports Not Found')
        for ep in excess_packages:
            excess_packages_report.add_row(ep.package_name, ', '.join(m for m in ep.modules))
        report.add_renderable(excess_packages_report)

    excess_modules = evaluation.module_report()
    if excess_modules:
        excess_modules_report = Table()
        excess_modules_report.add_column('Modules Used Without a Package')
        for em in excess_modules:
            excess_modules_report.add_row(em.name)
        report.add_renderable(excess_modules_report)

    return report


def format_extended_table_output(settings, evaluation):
    report = Table()
    report.add_column('Package')
    report.add_column('Used')
    report.add_column('Module')
    report.add_column('Used')
    all_modules = set(evaluation.modules)

    for package in evaluation.packages.values():
        mods = list(package.modules)
        if not mods:
            report.add_row(
                package.package_name, str(evaluation.evaluate_package(package.name)), '---', '---'
            )
        else:
            mods.sort()
            mod = mods.pop(0)
            all_modules.discard(mod)
            report.add_row(
                package.package_name,
                str(evaluation.evaluate_package(package.name)),
                mod,
                str(evaluation.evaluate_module(mod)),
            )
        for mod in mods:
            all_modules.discard(mod)
            report.add_row(
                '---',
                '---',
                mod,
                str(evaluation.evaluate_module(mod)),
            )

    this_project_modules = [mod for mod in all_modules if mod in settings.project_modules]
    if this_project_modules:
        this_project_modules.sort()
        mod = this_project_modules.pop(0)
        all_modules.discard(mod)
        report.add_row(
            settings.search_path, str(Confidence.SKIPPED), mod, str(evaluation.evaluate_module(mod))
        )
        for mod in this_project_modules:
            all_modules.discard(mod)
            report.add_row('---', '---', mod, str(evaluation.evaluate_module(mod)))

    stdlib_modules = [mod for mod in all_modules if mod in stdlib_module_names]
    if stdlib_modules:
        stdlib_modules.sort()
        mod = stdlib_modules.pop(0)
        all_modules.discard(mod)
        report.add_row('python', str(Confidence.SKIPPED), mod, str(evaluation.evaluate_module(mod)))
        for mod in stdlib_modules:
            all_modules.discard(mod)
            report.add_row('---', '---', mod, str(evaluation.evaluate_module(mod)))

    for mod in all_modules:
        report.add_row('???', '???', mod, str(evaluation.evaluate_module(mod)))
    return report



import sys

try:
    # more complete, but 3.10+
    stdlib_module_names = sys.stdlib_module_names
except AttributeError:
    # from python3.9
    stdlib_module_names = {
        '__future__',
        '_abc',
        '_aix_support',
        '_ast',
        '_asyncio',
        '_bisect',
        '_blake2',
        '_bootsubprocess',
        '_bz2',
        '_codecs',
        '_codecs_cn',
        '_codecs_hk',
        '_codecs_iso2022',
        '_codecs_jp',
        '_codecs_kr',
        '_codecs_tw',
        '_collections',
        '_collections_abc',
        '_compat_pickle',
        '_compression',
        '_contextvars',
        '_crypt',
        '_csv',
        '_ctypes',
        '_curses',
        '_curses_panel',
        '_datetime',
        '_dbm',
        '_decimal',
        '_elementtree',
        '_frozen_importlib',
        '_frozen_importlib_external',
        '_functools',
        '_gdbm',
        '_hashlib',
        '_heapq',
        '_imp',
        '_io',
        '_json',
        '_locale',
        '_lsprof',
        '_lzma',
        '_markupbase',
        '_md5',
        '_msi',
        '_multibytecodec',
        '_multiprocessing',
        '_opcode',
        '_operator',
        '_osx_support',
        '_overlapped',
        '_pickle',
        '_posixshmem',
        '_posixsubprocess',
        '_py_abc',
        '_pydecimal',
        '_pyio',
        '_queue',
        '_random',
        '_sha1',
        '_sha256',
        '_sha3',
        '_sha512',
        '_signal',
        '_sitebuiltins',
        '_socket',
        '_sqlite3',
        '_sre',
        '_ssl',
        '_stat',
        '_statistics',
        '_string',
        '_strptime',
        '_struct',
        '_symtable',
        '_thread',
        '_threading_local',
        '_tkinter',
        '_tracemalloc',
        '_uuid',
        '_warnings',
        '_weakref',
        '_weakrefset',
        '_winapi',
        '_zoneinfo',
        'abc',
        'aifc',
        'antigravity',
        'argparse',
        'array',
        'ast',
        'asynchat',
        'asyncio',
        'asyncore',
        'atexit',
        'audioop',
        'base64',
        'bdb',
        'binascii',
        'binhex',
        'bisect',
        'builtins',
        'bz2',
        'cProfile',
        'calendar',
        'cgi',
        'cgitb',
        'chunk',
        'cmath',
        'cmd',
        'code',
        'codecs',
        'codeop',
        'collections',
        'colorsys',
        'compileall',
        'concurrent',
        'configparser',
        'contextlib',
        'contextvars',
        'copy',
        'copyreg',
        'crypt',
        'csv',
        'ctypes',
        'curses',
        'dataclasses',
        'datetime',
        'dbm',
        'decimal',
        'difflib',
        'dis',
        'distutils',
        'doctest',
        'email',
        'encodings',
        'ensurepip',
        'enum',
        'errno',
        'faulthandler',
        'fcntl',
        'filecmp',
        'fileinput',
        'fnmatch',
        'formatter',
        'fractions',
        'ftplib',
        'functools',
        'gc',
        'genericpath',
        'getopt',
        'getpass',
        'gettext',
        'glob',
        'graphlib',
        'grp',
        'gzip',
        'hashlib',
        'heapq',
        'hmac',
        'html',
        'http',
        'idlelib',
        'imaplib',
        'imghdr',
        'imp',
        'importlib',
        'inspect',
        'io',
        'ipaddress',
        'itertools',
        'json',
        'keyword',
        'lib2to3',
        'linecache',
        'locale',
        'logging',
        'lzma',
        'mailbox',
        'mailcap',
        'marshal',
        'math',
        'mimetypes',
        'mmap',
        'modulefinder',
        'msilib',
        'msvcrt',
        'multiprocessing',
        'netrc',
        'nis',
        'nntplib',
        'nt',
        'ntpath',
        'nturl2path',
        'numbers',
        'opcode',
        'operator',
        'optparse',
        'os',
        'ossaudiodev',
        'parser',
        'pathlib',
        'pdb',
        'pickle',
        'pickletools',
        'pipes',
        'pkgutil',
        'platform',
        'plistlib',
        'poplib',
        'posix',
        'posixpath',
        'pprint',
        'profile',
        'pstats',
        'pty',
        'pwd',
        'py_compile',
        'pyclbr',
        'pydoc',
        'pydoc_data',
        'pyexpat',
        'queue',
        'quopri',
        'random',
        're',
        'readline',
        'reprlib',
        'resource',
        'rlcompleter',
        'runpy',
        'sched',
        'secrets',
        'select',
        'selectors',
        'shelve',
        'shlex',
        'shutil',
        'signal',
        'site',
        'smtpd',
        'smtplib',
        'sndhdr',
        'socket',
        'socketserver',
        'spwd',
        'sqlite3',
        'sre_compile',
        'sre_constants',
        'sre_parse',
        'ssl',
        'stat',
        'statistics',
        'string',
        'stringprep',
        'struct',
        'subprocess',
        'sunau',
        'symtable',
        'sys',
        'sysconfig',
        'syslog',
        'tabnanny',
        'tarfile',
        'telnetlib',
        'tempfile',
        'termios',
        'textwrap',
        'this',
        'threading',
        'time',
        'timeit',
        'tkinter',
        'token',
        'tokenize',
        'trace',
        'traceback',
        'tracemalloc',
        'tty',
        'turtle',
        'turtledemo',
        'types',
        'typing',
        'unicodedata',
        'unittest',
        'urllib',
        'uu',
        'uuid',
        'venv',
        'warnings',
        'wave',
        'weakref',
        'webbrowser',
        'winreg',
        'winsound',
        'wsgiref',
        'xdrlib',
        'xml',
        'xmlrpc',
        'zipapp',
        'zipfile',
        'zipimport',
        'zlib',
        'zoneinfo',
    }
    if sys.version_info.minor < 9:
        stdlib_module_names.remove('graphlib')
        stdlib_module_names.remove('zoneinfo')
    if sys.version_info.minor < 8:
        stdlib_module_names.remove('contextvars')
        stdlib_module_names.remove('dataclasses')
    if sys.version_info.minor < 7:
        stdlib_module_names.add('fpectl')
        stdlib_module_names.remove('secrets')
